unit DAutomatons;

interface

Const
  ConstMaxAutomatonDimention=33;
  ConstStrZeroState:String='-';

Type
  TStateSet=0..ConstMaxAutomatonDimention;
  TMealyState=Record  //      ()
    NextState,          // 
    OutSignal:TStateSet;// 
  End;
  TMealyStates=Array[1..ConstMaxAutomatonDimention, 1..ConstMaxAutomatonDimention] Of TMealyState; //    ()
  TMooreStates=Array[0..ConstMaxAutomatonDimention, 1..ConstMaxAutomatonDimention] Of TStateSet;   //    ()
  TMinimizeTable=TMooreStates;
  TOutsByZ=Array[1..ConstMaxAutomatonDimention] Of TStateSet;

  TStateGroups=Object  //      
  Protected
    Groups:Array[1..ConstMaxAutomatonDimention] Of TMealyState;
    _Count:TStateSet;
  Public
    Procedure Clear;
    Function Add(a, w:TStateSet):Integer;
    Function Find(a, w:TStateSet):Integer;
    Function Get(i:TStateSet; Var a, w:TStateSet):Integer;
    Function Count:Integer;
  End;

  TOutGroups=Object  //    
  Protected
    Groups:Array[1..ConstMaxAutomatonDimention] Of TOutsByZ;
    _Count:TStateSet;
    _ZCount:TStateSet;

    Function Compare(Var Outs1, Outs2:TOutsByZ):Boolean;
  Public
    Procedure Clear(Z:TStateSet);
    Function Add(Var Outs:TOutsByZ):Integer;
    Function Find(Var Outs:TOutsByZ):Integer;
    Function Get(i:TStateSet; Var Outs:TOutsByZ):Integer;
    Function GetZCount:TStateSet;
    Function Count:Integer;
  End;

  TAutomaton=Object
  Public
    LastError:Integer;
  Protected
    Defined:Boolean;
    StateCount_A,               //  (A)
    InputCount_Z,               //   (Z)
    OutCount_W:TStateSet;       //   (W)
    CurrentState:TStateSet;     //  

  Protected
    Procedure InternalInit; Virtual;

  Public
    Constructor Init;
    Destructor Done; Virtual;

    Function CheckState_A(a:TStateSet):Boolean;
    Function CheckInput_Z(z:TStateSet):Boolean;
    Function CheckOutput_W(w:TStateSet):Boolean;
    
    Procedure Clear; Virtual;

    Function GetACount:TStateSet;
    Function GetZCount:TStateSet;
    Function GetWCount:TStateSet;

    Function GetCurrentState:TStateSet;
    Function SetCurrentState(a:TStateSet):Integer;

    Function GetOut(z:TStateSet):TStateSet; Virtual; Abstract;                  //  
    Function GetOutByState(a, z:TStateSet):TStateSet; Virtual; Abstract;        // --//--   
    Function GetNextState(z:TStateSet):TStateSet; Virtual; Abstract;            //  
    Function GetNextStateByState(a, z:TStateSet):TStateSet; Virtual; Abstract;  // --//--   
    Function GetStateInfo(z:TStateSet; Var NextA, w:TStateSet):Integer; Virtual; Abstract;            //    
    Function GetStateInfoByState(a, z:TStateSet; Var NextA, w:TStateSet):Integer; Virtual; Abstract;  // --//--   

    Function ReDefine(Var A, Z, W:TStateSet):Integer; Virtual;

    Function ProcessStep(z:TStateSet; Var NextA, w:TStateSet):Integer; Virtual; Abstract;

    Function Minimize:Integer; Virtual; Abstract;
  End;

  TMealyAutomaton=Object(TAutomaton)      // 
  Protected
    States:TMealyStates;        //

  Public
    Constructor Init;
    Destructor Done; Virtual;

    Procedure Clear; Virtual;

    Function GetOut(z:TStateSet):TStateSet; Virtual;                  //  
    Function GetOutByState(a, z:TStateSet):TStateSet; Virtual;        // --//--   
    Function GetNextState(z:TStateSet):TStateSet; Virtual;            //  
    Function GetNextStateByState(a, z:TStateSet):TStateSet; Virtual;  // --//--   
    Function GetStateInfo(z:TStateSet; Var NextA, w:TStateSet):Integer; Virtual;            //    
    Function GetStateInfoByState(a, z:TStateSet; Var NextA, w:TStateSet):Integer; Virtual;  // --//--   

    Function InsertState(a, z, NextA, w:TStateSet):Integer;

    Function ProcessStep(z:TStateSet; Var NextA, w:TStateSet):Integer; Virtual;

    Function Minimize:Integer; Virtual;
  End;

  TMooreAutomaton=Object(TAutomaton)      // 
  Protected
    States:TMooreStates;        //

  Public
    Constructor Init;
    Destructor Done; Virtual;

    Procedure Clear; Virtual;

    Function GetOut(z:TStateSet):TStateSet; Virtual;                  //  
    Function GetOutByState(a, z:TStateSet):TStateSet; Virtual;        // --//--   
    Function GetNextState(z:TStateSet):TStateSet; Virtual;            //  
    Function GetNextStateByState(a, z:TStateSet):TStateSet; Virtual;  // --//--   
    Function GetStateInfo(z:TStateSet; Var NextA, w:TStateSet):Integer; Virtual;            //    
    Function GetStateInfoByState(a, z:TStateSet; Var NextA, w:TStateSet):Integer; Virtual;  // --//--   

    Function InsertOut(a, w:TStateSet):Integer;
    Function InsertState(a, z, NextA, w:TStateSet):Integer;

    Function ProcessStep(z:TStateSet; Var NextA, w:TStateSet):Integer; Virtual;

    Function Minimize:Integer; Virtual;
  End;

  Function ConvertMealyToMoore(Var MealyA:TMealyAutomaton; Var MooreA:TMooreAutomaton):Integer;
  Function ConvertMooreToMealy(Var MooreA:TMooreAutomaton; Var MealyA:TMealyAutomaton):Integer;

implementation

Function CreateStateGroup(Var MealyA:TMealyAutomaton; Var StateGroup:TStateGroups):Integer; Forward;
Function MinimizeTableNewGroups(Var MinTable:TMinimizeTable; Var GroupCnt:TStateSet; ACnt, ZCnt:TStateSet):Integer; Forward;
Function MinimizeTableNewOuts(Var MinTable:TMinimizeTable; ACnt, ZCnt:TStateSet):Integer; Forward;
Function CompareMinimizeTables(Var MinTable1, MinTable2:TMinimizeTable; ACnt, ZCnt:TStateSet):Boolean; Forward;

Function TruncateMealyStates(Var States:TMealyStates; Var ACnt:TStateSet; ZCnt:TStateSet):Integer; Forward;
Function TruncateMooreStates(Var States:TMooreStates; Var ACnt:TStateSet; ZCnt:TStateSet):Integer; Forward;

Function TOutGroups.Compare(Var Outs1, Outs2:TOutsByZ):Boolean;
Var
  i:Integer;
Begin
  Result:=False;
  For i:=1 To _ZCount Do
  Begin
    If Outs1[i]<>Outs2[i] Then Exit;
  End;
  Result:=True;
End;

Procedure TOutGroups.Clear(Z:TStateSet);
Begin
  FillChar(Groups, SizeOf(Groups), 0);
  _Count:=0;
  _ZCount:=Z;
End;

Function TOutGroups.Add(Var Outs:TOutsByZ):Integer;
Begin
  Result:=Find(Outs);
  If Result>0 Then Exit;
  If _Count>=ConstMaxAutomatonDimention Then
  Begin
    Result:=-1;
    Exit;
  End;

  Inc(_Count);
  Groups[_Count]:=Outs;
  Result:=_Count;
End;

Function TOutGroups.Find(Var Outs:TOutsByZ):Integer;
Var
  i:Integer;
Begin
  Result:=-1;
  If _Count<=0 Then Exit;
  For i:=1 To _Count Do
  Begin
    If Compare(Outs, Groups[i]) Then
    Begin
      Result:=i;
      Exit;
    End;
  End;
End;

Function TOutGroups.Get(i:TStateSet; Var Outs:TOutsByZ):Integer;
Begin
  Result:=-1;
  If i>_Count Then Exit;
  Outs:=Groups[i];
  Result:=0;
End;

Function TOutGroups.GetZCount:TStateSet;
Begin
  Result:=_ZCount;
End;

Function TOutGroups.Count:Integer;
Begin
  Result:=_Count;
End;


Procedure TStateGroups.Clear;
Begin
  FillChar(Groups, SizeOf(Groups), 0);
  _Count:=0;
End;

Function TStateGroups.Add(a, w:TStateSet):Integer;
Begin
  Result:=Find(a, w);
  If Result>0 Then Exit;
  If _Count>=ConstMaxAutomatonDimention Then
  Begin
    Result:=-1;
    Exit;
  End;

  Inc(_Count);
  Groups[_Count].NextState:=a;
  Groups[_Count].OutSignal:=w;
  Result:=_Count;
End;

Function TStateGroups.Find(a, w:TStateSet):Integer;
Var
  i:Integer;
Begin
  Result:=-1;
  If _Count<=0 Then Exit;

  For i:=1 To _Count Do
  Begin
    If ((a=Groups[i].NextState) And (w=Groups[i].OutSignal)) Then
    Begin
      Result:=i;
      Exit;
    End;
  End;
  Result:=-1;
End;

Function TStateGroups.Get(i:TStateSet; Var a, w:TStateSet):Integer;
Begin
  If i>_Count Then
  Begin
    Result:=-1;
    Exit;
  End;
  a:=Groups[i].NextState;
  w:=Groups[i].OutSignal;
  Result:=0;
End;

Function TStateGroups.Count:Integer;
Begin
  Result:=_Count;
End;


Procedure TAutomaton.InternalInit;
Begin
  Clear;
  Defined:=False;
  StateCount_A:=0;
  InputCount_Z:=0;
  OutCount_W:=0;
  CurrentState:=0;
  LastError:=0;
End;

Function TAutomaton.CheckState_A(a:TStateSet):Boolean;
Begin
  Result:=((a>=1) And (a<=StateCount_A));
End;
Function TAutomaton.CheckInput_Z(z:TStateSet):Boolean;
Begin
  Result:=((z>=1) And (z<=InputCount_Z));
End;
Function TAutomaton.CheckOutput_W(w:TStateSet):Boolean;
Begin
  Result:=((w>=1) And (w<=OutCount_W));
End;

Constructor TAutomaton.Init;
Begin
  InternalInit;
End;

Destructor TAutomaton.Done;
Begin
  InternalInit;
End;

Procedure TAutomaton.Clear;
Begin
  If Defined Then CurrentState:=1 Else CurrentState:=0;
End;

Function TAutomaton.GetACount:TStateSet;
Begin
  Result:=StateCount_A;
End;

Function TAutomaton.GetZCount:TStateSet;
Begin
  Result:=InputCount_Z;
End;

Function TAutomaton.GetWCount:TStateSet;
Begin
  Result:=OutCount_W;
End;

Function TAutomaton.GetCurrentState:TStateSet;
Begin
  Result:=CurrentState;
  If CurrentState<1 Then LastError:=-1;
End;

Function TAutomaton.SetCurrentState(a:TStateSet):Integer;
Begin
  If Not(CheckState_A(a)) Then
  Begin
    Result:=-1;
    LastError:=-1;
    Exit;
  End;
  CurrentState:=a;
  Result:=0;
  LastError:=0;
End;

Function TAutomaton.ReDefine(Var A, Z, W:TStateSet):Integer;
Begin
  Result:=0;
  If ((A<1) Or (A>ConstMaxAutomatonDimention)) Then
  Begin
    A:=1;
    Result:=-1;
  End;
  If ((Z<1) Or (Z>ConstMaxAutomatonDimention)) Then
  Begin
    W:=1;
    Result:=-1;
  End;
  If ((W<1) Or (W>ConstMaxAutomatonDimention)) Then
  Begin
    W:=1;
    Result:=-1;
  End;
  If Result<0 Then      //  ,   
  Begin
    LastError:=-1;
    Exit;
  End;

  StateCount_A:=A;
  InputCount_Z:=Z;
  OutCount_W:=W;
  CurrentState:=1;

  Defined:=True;
  Result:=0;
  LastError:=0;
End;


Constructor TMealyAutomaton.Init;
Begin
  InternalInit;
End;

Destructor TMealyAutomaton.Done;
Begin
  InternalInit;
End;

Procedure TMealyAutomaton.Clear;
Begin
  inherited Clear;
  FillChar(States, SizeOf(States), 0);
End;

Function TMealyAutomaton.GetOut(z:TStateSet):TStateSet;
Begin
  If Not(CheckInput_Z(z)) Then
  Begin
    Result:=0;
    LastError:=-1;
    Exit;
  End;

  Result:=States[z, CurrentState].OutSignal;
  LastError:=0;
End;

Function TMealyAutomaton.GetOutByState(a, z:TStateSet):TStateSet;
Begin
  If Not(CheckState_A(a)) Or Not(CheckInput_Z(z)) Then
  Begin
    Result:=0;
    LastError:=-1;
    Exit;
  End;
  Result:=States[z, a].OutSignal;
  LastError:=0;
End;

Function TMealyAutomaton.GetNextState(z:TStateSet):TStateSet;
Begin
  If Not(CheckInput_Z(z)) Then
  Begin
    Result:=0;
    LastError:=-1;
    Exit;
  End;

  Result:=States[z, CurrentState].NextState;
  LastError:=0;
End;

Function TMealyAutomaton.GetNextStateByState(a, z:TStateSet):TStateSet;
Begin
  If Not(CheckState_A(a)) Or Not(CheckInput_Z(z)) Then
  Begin
    Result:=0;
    LastError:=-1;
    Exit;
  End;

  Result:=States[z, a].NextState;
  LastError:=0;
End;

Function TMealyAutomaton.GetStateInfo(z:TStateSet; Var NextA, w:TStateSet):Integer;            //    
Begin
  Result:=GetStateInfoByState(CurrentState, z, NextA, W);
End;

Function TMealyAutomaton.GetStateInfoByState(a, z:TStateSet; Var NextA, w:TStateSet):Integer;  // --//--   
Begin
  Result:=0;
  NextA:=GetNextStateByState(a, z);
  If LastError<0 Then
  Begin
    Result:=-1;
    Exit;
  End;
  w:=GetOutByState(a, z);
  If LastError<0 Then Result:=-1;
End;

Function TMealyAutomaton.InsertState(a, z, NextA, w:TStateSet):Integer;
Begin
  If Not(CheckState_A(a)) Or Not(CheckInput_Z(z)) Then
  Begin
    Result:=-1;
    LastError:=-1;
    Exit;
  End;
  If ((NextA>StateCount_A) Or (w>OutCount_W)) Then  //   
  Begin
    Result:=-1;
    LastError:=-1;
    Exit;
  End;
{
  If ((NextA=0) Or (w=0)) Then  //   ,    :)
  Begin
    NextA:=0;
    w:=0;
  End;
}

  States[z, a].NextState:=NextA;
  States[z, a].OutSignal:=w;

  Result:=0;
  LastError:=0;
End;

Function TMealyAutomaton.ProcessStep(z:TStateSet; Var NextA, w:TStateSet):Integer;
Begin
  Result:=GetStateInfo(z, NextA, w);        //    
  If NextA<>0 Then SetCurrentState(NextA);  //   ,    
End;

Function TMealyAutomaton.Minimize:Integer;
Var
  NewStates:TMealyStates;
  PrevMinTable, NextMinTable:TMinimizeTable;
  a, z:TStateSet;
  Group, GroupCount:TStateSet;
Begin
  Result:=-1;
  If Not(Defined) Then Exit;

 {  }
  FillChar(NextMinTable, SizeOf(NextMinTable), 0);
  For a:=1 To StateCount_A Do
  Begin
    NextMinTable[0, a]:=1;
  End;
  For a:=1 To StateCount_A Do
  Begin
    For z:=1 To InputCount_Z Do
    Begin
      NextMinTable[z, a]:=GetOutByState(a, z);
    End;
  End;
  GroupCount:=1;
 {    }
  Repeat
    PrevMinTable:=NextMinTable;
   {  }
    MinimizeTableNewGroups(NextMinTable, GroupCount, StateCount_A, InputCount_Z);
   {     }
    For a:=1 To StateCount_A Do
    Begin
      For z:=1 To InputCount_Z Do
      Begin
        NextMinTable[z, a]:=GetNextStateByState(a, z);
       {  ,     }
        If NextMinTable[z, a]=0 Then
          NextMinTable[z, a]:=a;
      End;
    End;
   {  }
    MinimizeTableNewOuts(NextMinTable, StateCount_A, InputCount_Z);
  Until CompareMinimizeTables(PrevMinTable, NextMinTable, StateCount_A, InputCount_Z);

 {  }
  FillChar(NewStates, SizeOf(NewStates), 0);
  For a:=1 To StateCount_A Do  //    
  Begin
    Group:=NextMinTable[0, a];
    For z:=1 To InputCount_Z Do
    Begin
      NewStates[z, Group].NextState:=NextMinTable[z, a];
      NewStates[z, Group].OutSignal:=GetOutByState(a, z);
    End;
  End;
 { 
   ,    }
  TruncateMealyStates(NewStates, GroupCount, InputCount_Z);

  ReDefine(GroupCount, InputCount_Z, OutCount_W);
  States:=NewStates;
  
  Result:=0;
  LastError:=0;
End;


Constructor TMooreAutomaton.Init;
Begin
  InternalInit;
End;

Destructor TMooreAutomaton.Done;
Begin
  InternalInit;
End;

Procedure TMooreAutomaton.Clear;
Begin
  inherited Clear;
  FillChar(States, SizeOf(States), 0);
End;

Function TMooreAutomaton.GetOut(z:TStateSet):TStateSet;
Begin
  Result:=GetOutByState(CurrentState, z);
End;

Function TMooreAutomaton.GetOutByState(a, z:TStateSet):TStateSet;
Var
  NextA:TStateSet;
Begin
  LastError:=0;
  If (Not(CheckState_A(a)) Or (z>ConstMaxAutomatonDimention)) Then
  Begin
    Result:=0;
    LastError:=-1;
    Exit;
  End;
 {    ,   }
  If z=0 Then
  Begin
    Result:=States[0, a];
    LastError:=0;
    Exit;
  End;

 {   }
  NextA:=GetNextStateByState(a, z);
  If NextA=0 Then
    Result:=States[0, a]  //   ,     
  Else
    Result:=States[0, NextA];  // - 
End;

Function TMooreAutomaton.GetNextState(z:TStateSet):TStateSet;
Begin
  Result:=GetNextStateByState(CurrentState, z);
End;

Function TMooreAutomaton.GetNextStateByState(a, z:TStateSet):TStateSet;
Begin
  If Not(CheckState_A(a)) Or Not(CheckInput_Z(z)) Then
  Begin
    Result:=0;
    LastError:=-1;
    Exit;
  End;

  Result:=States[z, a];
  LastError:=0;
End;

Function TMooreAutomaton.GetStateInfo(z:TStateSet; Var NextA, w:TStateSet):Integer;            //    
Begin
  Result:=GetStateInfoByState(CurrentState, z, NextA, W);
End;

Function TMooreAutomaton.GetStateInfoByState(a, z:TStateSet; Var NextA, w:TStateSet):Integer;  // --//--   
Begin
  Result:=0;
  NextA:=GetNextStateByState(a, z);
  If LastError<0 Then
  Begin
    Result:=-1;
    Exit;
  End;
  w:=GetOutByState(a, z);
  If LastError<0 Then Result:=-1;
End;

Function TMooreAutomaton.InsertOut(a, w:TStateSet):Integer;
Begin
  If Not(CheckState_A(a)) Then
  Begin
    Result:=-1;
    LastError:=-1;
    Exit;
  End;
  If w>OutCount_W Then  //   
  Begin
    Result:=-1;
    LastError:=-1;
    Exit;
  End;

  States[0, a]:=w;

  Result:=0;
  LastError:=0;
End;

Function TMooreAutomaton.InsertState(a, z, NextA, w:TStateSet):Integer;
Begin
  If Not(CheckState_A(a)) Or Not(CheckInput_Z(z)) Then
  Begin
    Result:=-1;
    LastError:=-1;
    Exit;
  End;
  If (NextA>StateCount_A) Then  //   
  Begin
    Result:=-1;
    LastError:=-1;
    Exit;
  End;

  States[z, a]:=NextA;

  Result:=0;
  LastError:=0;
End;

Function TMooreAutomaton.ProcessStep(z:TStateSet; Var NextA, w:TStateSet):Integer;
Begin
  Result:=GetStateInfo(z, NextA, w);        //    
  If NextA<>0 Then SetCurrentState(NextA);  //   ,    
End;

Function TMooreAutomaton.Minimize:Integer;
Var
  NewStates:TMooreStates;
  PrevMinTable, NextMinTable:TMinimizeTable;
  a, z, NewOut:TStateSet;
  Group, GroupCount:TStateSet;
  Groups:Array[0..ConstMaxAutomatonDimention] Of TStateSet;
Begin
  Result:=-1;
  If Not(Defined) Then Exit;

 {  }
  GroupCount:=0;
  FillChar(Groups, SizeOf(Groups), 0);
  FillChar(NextMinTable, SizeOf(NextMinTable), 0);
  For a:=1 To StateCount_A Do
  Begin
    NewOut:=GetOutByState(a, 0);
    If Groups[NewOut]=0 Then
    Begin
      Inc(GroupCount);
      Groups[NewOut]:=GroupCount;
//      NextMinTable[0, a]:=GroupCount;
    End;
    NextMinTable[0, a]:=Groups[NewOut];
{     //fixed 13.06.03
    NextMinTable[0, a]:=GetOutByState(a, 0);
    If NextMinTable[0, a]=0 Then
    Begin
      GroupCount:=OutCount_W+1;
      NextMinTable[0, a]:=GroupCount;
    End;
}
  End;
  For a:=1 To StateCount_A Do
  Begin
    For z:=1 To InputCount_Z Do
    Begin
      NextMinTable[z, a]:=GetOutByState(a, z);
    End;
  End;
 {    }
  Repeat
    PrevMinTable:=NextMinTable;
   {  }
    MinimizeTableNewGroups(NextMinTable, GroupCount, StateCount_A, InputCount_Z);
   {     }
    For a:=1 To StateCount_A Do
    Begin
      For z:=1 To InputCount_Z Do
      Begin
        NextMinTable[z, a]:=GetNextStateByState(a, z);
       {  ,     }
        If NextMinTable[z, a]=0 Then
          NextMinTable[z, a]:=a;
      End;
    End;
   {  }
    MinimizeTableNewOuts(NextMinTable, StateCount_A, InputCount_Z);
  Until CompareMinimizeTables(PrevMinTable, NextMinTable, StateCount_A, InputCount_Z);

 {  }
  FillChar(NewStates, SizeOf(NewStates), 0);
  For a:=1 To StateCount_A Do  //    
  Begin
    Group:=NextMinTable[0, a];
    NewStates[0, Group]:=GetOutByState(a, 0);
    For z:=1 To InputCount_Z Do
    Begin
      NewStates[z, Group]:=NextMinTable[z, a];
    End;
  End;
 { 
   ,    }
  TruncateMooreStates(NewStates, GroupCount, InputCount_Z);

  ReDefine(GroupCount, InputCount_Z, OutCount_W);
  States:=NewStates;
  
  Result:=0;
  LastError:=0;
End;

Function ConvertMealyToMoore(Var MealyA:TMealyAutomaton; Var MooreA:TMooreAutomaton):Integer;
Var
  StateGroup:TStateGroups;
  a, z, w:TStateSet;
  ACnt, ZCnt, WCnt:TStateSet;
  Tmp, TmpA, TmpW:TStateSet;
  NextB:TStateSet;
  NowB:Integer;
  UnDef:Boolean;
Begin
  Result:=0;
  If Not(MealyA.Defined) Then  //    ,  
  Begin
    Result:=-1;
    Exit;
  End;
 { }
  If CreateStateGroup(MealyA, StateGroup)<0 Then  //    
  Begin
   Result:=-1;
   Exit;
  End;

 {  }
  ACnt:=StateGroup.Count;
  ZCnt:=MealyA.InputCount_Z;
  WCnt:=MealyA.OutCount_W;
  MooreA.Init;
  MooreA.ReDefine(ACnt, ZCnt, WCnt);
 {  }
  For a:=1 To ACnt Do
  Begin
    StateGroup.Get(a, Tmp, w);
    MooreA.InsertOut(a, w);
  End;
 {    }
  For a:=1 To MealyA.StateCount_A Do
  Begin
    For z:=1 To ZCnt Do
    Begin
      UnDef:=False;
      MealyA.GetStateInfoByState(a, z, TmpA, TmpW);
     { }
      If ((TmpA=0) And (TmpW=0)) Then
      Begin
        TmpA:=a;  //  
        TmpW:=0;
      End;
     { }
      If ((TmpA=0) And (TmpW<>0)) Then
      Begin
        TmpA:=a;  //  
        UnDef:=True;
      End;
      NextB:=StateGroup.Find(TmpA, TmpW);  // 
     {  ,     }
      For w:=0 To WCnt Do
      Begin
       {  ,   }
        NowB:=StateGroup.Find(a, w);
        If NowB>0 Then
        Begin
          MooreA.InsertState(NowB, z, NextB, 0);
         { }
          If ((NextB=NowB) And ((w=0) Or UnDef)) Then  //   
          Begin
            MooreA.InsertState(NowB, z, 0, 0);
          End;
        End;
      End;
    End;
  End;
 { DONE 5 -oDarkick :    ConvertMealyToMoore! }
End;

Function ConvertMooreToMealy(Var MooreA:TMooreAutomaton; Var MealyA:TMealyAutomaton):Integer;
Var
  a, z, w, NextA:TStateSet;
  ACnt, ZCnt, WCnt:TStateSet;
Begin
  If Not(MooreA.Defined) Then  //    ,  
  Begin
    Result:=-1;
    Exit;
  End;

  ACnt:=MooreA.GetACount;
  ZCnt:=MooreA.GetZCount;
  WCnt:=MooreA.GetWCount;
  MealyA.Init;  //   (  )
  MealyA.ReDefine(ACnt, ZCnt, WCnt);

  For z:=1 To ZCnt Do
  Begin
    For a:=1 To ACnt Do
    Begin
      NextA:=MooreA.GetNextStateByState(a, z);
      w:=MooreA.GetOutByState(a, z);
      MealyA.InsertState(a, z, NextA, w);
    End;
  End;
  Result:=0;
End;


Function CreateStateGroup(Var MealyA:TMealyAutomaton; Var StateGroup:TStateGroups):Integer;
Var
  aa, ww, a, z, w:TStateSet;
  ACnt, ZCnt, WCnt:TStateSet;
  NextA:TStateSet;
  StateExists:Boolean;
Begin
  StateGroup.Clear;

  ACnt:=MealyA.StateCount_A;
  ZCnt:=MealyA.InputCount_Z;
  WCnt:=MealyA.OutCount_W;

  For aa:=1 To ACnt Do
  Begin
    StateExists:=False;
    For ww:=0 To WCnt Do
    Begin
      For z:=1 To ZCnt Do    // 
      Begin                  //
        For a:=1 To ACnt Do  // 
        Begin
          NextA:=MealyA.GetNextStateByState(a, z);
          w:=MealyA.GetOutByState(a, z);
          If ((NextA=aa) And (w=ww)) Then  //   
          Begin
            StateExists:=True;  //   aa     
            If StateGroup.Add(aa, w)<0 Then
            Begin  // !
              Result:=-1;
              Exit;
            End;
          End;
        End;
      End;
    End;
    If Not(StateExists) Then //    aa    
    Begin
      If StateGroup.Add(aa, 0)<0 Then
      Begin  // !
        Result:=-1;
        Exit;
      End;
    End;
  End;

 {   }
 {( )}
 {      (a, 0)}
  For z:=1 To ZCnt Do    // 
  Begin                  //
    For a:=1 To ACnt Do  // 
    Begin
      NextA:=MealyA.GetNextStateByState(a, z);
      w:=MealyA.GetOutByState(a, z);
      If ((NextA=0) And (w=0)) Then  //   
      Begin
        If StateGroup.Add(a, 0)<0 Then
        Begin  // !
          Result:=-1;
          Exit;
        End;
      End;
    End;
  End;
 {   }
 {( ,   )}
 {      }
  For z:=1 To ZCnt Do    // 
  Begin                  //
    For a:=1 To ACnt Do  // 
    Begin
      NextA:=MealyA.GetNextStateByState(a, z);
      w:=MealyA.GetOutByState(a, z);
      If ((NextA=0) And (w<>0)) Then  //    
      Begin
        If StateGroup.Add(a, w)<0 Then
        Begin  // !
          Result:=-1;
          Exit;
        End;
      End;
    End;
  End;

  Result:=0;
End;

{   }
Function MinimizeTableNewGroups(Var MinTable:TMinimizeTable; Var GroupCnt:TStateSet; ACnt, ZCnt:TStateSet):Integer;
Var
  PrevMinTable:TMinimizeTable;
  OutGroups:TOutGroups;
  Outs:TOutsByZ;
  a, z:Integer;
  i, NewGroup, NewGroupCnt:Integer;
Begin
//  Result:=-1;
  PrevMinTable:=MinTable;
  NewGroupCnt:=0;

  For i:=1 To GroupCnt Do  //   
  Begin
    OutGroups.Clear(ZCnt);
    For a:=1 To ACnt Do  //   
    Begin
      If PrevMinTable[0, a]=i Then  //     
      Begin
        For z:=1 To ZCnt Do  //  
        Begin
          Outs[z]:=PrevMinTable[z, a];
        End;
        NewGroup:=OutGroups.Add(Outs);  //   
        MinTable[0, a]:=NewGroupCnt+NewGroup;
      End;
    End;
    Inc(NewGroupCnt, OutGroups.Count);
  End;
  GroupCnt:=NewGroupCnt;
  Result:=0;
End;
{  }
Function MinimizeTableNewOuts(Var MinTable:TMinimizeTable; ACnt, ZCnt:TStateSet):Integer;
Var
  a, z:Integer;
  NextA, NextOut:TStateSet;
Begin
//  Result:=-1;

  For a:=1 To ACnt Do  //   
  Begin
    For z:=1 To ZCnt Do
    Begin  //  
      NextA:=MinTable[z, a];
      NextOut:=MinTable[0, NextA];
      MinTable[z, a]:=NextOut;
    End;
  End;

  Result:=0;
End;

Function CompareMinimizeTables(Var MinTable1, MinTable2:TMinimizeTable; ACnt, ZCnt:TStateSet):Boolean;
Var
  a, z:TStateSet;
Begin
  Result:=False;
  For z:=0 To ZCnt Do
  Begin
    For a:=1 To ACnt Do
    Begin
      If MinTable1[z, a]<>MinTable2[z, a] Then Exit;
    End;
  End;
  Result:=True;
End;

Function TruncateMealyStates(Var States:TMealyStates; Var ACnt:TStateSet; ZCnt:TStateSet):Integer;
Var
  StatesUsage:Array[1..ConstMaxAutomatonDimention] Of TStateSet;
  a, z, aa:Integer;
  NextA:TStateSet;
Begin
  Result:=0;
  If ACnt<2 Then Exit;
  FillChar(StatesUsage, SizeOf(StatesUsage), 0);
 { ,    }
  For a:=1 To ACnt Do
  Begin
    For z:=1 To ZCnt Do
    Begin
      NextA:=States[z, a].NextState;
      If NextA<>a Then StatesUsage[NextA]:=1;
    End;
  End;
 {  }
  For aa:=ACnt DownTo 2 Do
  Begin
    If StatesUsage[aa]=0 Then
    Begin
     {  }
      For a:=1 To ACnt Do
      Begin
        For z:=1 To ZCnt Do
        Begin
          NextA:=States[z, a].NextState;
          If NextA>aa Then Dec(States[z, a].NextState);
        End;
      End;
     { }
      For a:=aa To ACnt Do
      Begin
        For z:=1 To ZCnt Do
        Begin
          States[z, a]:=States[z, a+1];
        End;
      End;
      Dec(ACnt);
      Result:=aa;
      Break;
    End;
  End;
  If Result>0 Then
    Result:=TruncateMealyStates(States, ACnt, ZCnt);
End;

Function TruncateMooreStates(Var States:TMooreStates; Var ACnt:TStateSet; ZCnt:TStateSet):Integer;
Var
  StatesUsage:Array[1..ConstMaxAutomatonDimention] Of TStateSet;
  a, z, aa:Integer;
  NextA:TStateSet;
Begin
  Result:=0;
  If ACnt<2 Then Exit;
  FillChar(StatesUsage, SizeOf(StatesUsage), 0);
 { ,    }
  For a:=1 To ACnt Do
  Begin
    For z:=1 To ZCnt Do
    Begin
      NextA:=States[z, a];
      If NextA<>a Then StatesUsage[NextA]:=1;
    End;
  End;
 {  }
  For aa:=ACnt DownTo 2 Do
  Begin
    If StatesUsage[aa]=0 Then
    Begin
     {  }
      For a:=1 To ACnt Do
      Begin
        For z:=1 To ZCnt Do
        Begin
          NextA:=States[z, a];
          If NextA>aa Then Dec(States[z, a]);
        End;
      End;
     { }
      For a:=aa To ACnt Do
      Begin
        For z:=0 To ZCnt Do
        Begin
          States[z, a]:=States[z, a+1];
        End;
      End;
      Dec(ACnt);
      Result:=aa;
      Break;
    End;
  End;
  If Result>0 Then
    Result:=TruncateMooreStates(States, ACnt, ZCnt);
End;

{ DONE 3 -oDarkick :       CreateStateGroup }
{ DONE 2 -oDarkick : :         ,  
:   }
{ DONE 2  -oDarkick : :      (  - ,   )
:  }
{ DONE 2 -oDarkick : :          ()   
:  .    ,   
:     TMoore.InsertState}

{ TODO 5 -oDarkick : !:     . (   ) }
{ TODO 1 -oDarkick :
:          
}
{ TODO 5 :     .      }

end.

